#!/usr/bin/python2
# -*- coding: utf-8 -*-

import os
import sqlite3 as lite
from decimal import Decimal

try:
    from process.error_handling import KillByUser
except ImportError:
    from trifusion.process.error_handling import KillByUser


VAR_LENGTH = 0
VAR_TAXON = 1
cur = None

"""
Read all fasta files from a folder, placing the genes present on those fasta
into a single variable
"""


def get_genes(files_dir):

    genes = {}

    # Filter hidden files and not fasta files (from extension)
    files_list = [x for x in os.listdir(files_dir)
                  if not x.startswith(".") or
                  x.endswith(".fasta") or
                  x.endswith(".fas") or
                  x.endswith(".fa")]

    # for all fast files in fasta directory
    for fasta in files_list:

        # get taxon from file name
        # splitted = fasta.split(".")
        taxon = os.path.splitext(fasta)[0]

        # open file
        fasta_file = open(os.path.join(files_dir, fasta), "r")

        gene = ''
        length = 0
        for line in fasta_file:
            # clean '\n'
            line = line.strip()
            # ignore empty lines and stop codons
            if not line: # or line.endswith("*"):
                continue

            new_gene = True if line.startswith(">") else False

            if new_gene:
                # save previous gene info
                if gene:
                    genes[gene][VAR_LENGTH] = length

                # save new gene info
                gene = line[1:]
                genes[gene] = [None, taxon]

                # reset vars
                length = 0
            else:
                length += len(line)

        genes[gene][VAR_LENGTH] = length

        fasta_file.close()

    return genes


def get_taxon_and_length(subject, genes):

    subject["queryTaxon"] = genes[subject["queryId"]][VAR_TAXON]
    subject["subjectTaxon"] = genes[subject["subjectId"]][VAR_TAXON]
    subject["queryLength"] = genes[subject["queryId"]][VAR_LENGTH]
    subject["subjectLength"] = genes[subject["subjectId"]][VAR_LENGTH]

    try:
        subject["subjectTaxon"]
    except KeyError:
        print ("couldn't find taxon for gene " + subject["subjectId"])
    try:
        subject["queryTaxon"]
    except KeyError:
        print ("couldn't find taxon for gene " + subject["queryId"])

    return subject["queryLength"] < subject["subjectLength"]


def print_previous_subject(subject, db):

    non_overlap = non_overlapping_match(subject)

    percent_ident = '{0:.4g}'.format(
        int((float(subject["totalIdentities"]) /
         float(subject["totalLength"]) * 10 + .5)) / 10)
    
    shorter_length = subject["queryLength"] if subject["queryShorter"] \
        else subject["subjectLength"]

    percent_match = '{0:.3g}'.format((float(non_overlap) /
                                      float(shorter_length) * 1000 + .5) / 10)

    db.execute("INSERT OR IGNORE INTO SimilarSequences VALUES(?, ?, ?, ?, ?, ? ,?, ?)",
                [subject["queryId"],
                subject["subjectId"],
                subject["queryTaxon"],
                subject["subjectTaxon"],
                float(subject["evalueMant"]),
                int(subject["evalueExp"]),
                float(percent_ident),
                float(percent_match)])
    

def format_evalue(evalue):
    if evalue == '0':
        return 0, 0
    evalue = "{:.2E}".format(Decimal(evalue))
    if evalue.startswith('e'):
        evalue = '1' + evalue

    return [round(float(x), 2) for x in evalue.split("E")]


def non_overlapping_match(subject):

    # flatten lists
    hsps = subject["hspspans"]
    hsps.sort(key=lambda x: int(x[0]))
    original = hsps.pop(0)

    if not original:
        return 0

    start = 0
    end = 1

    original = get_start_end(original)
    length = 0

    for h in hsps:
        h = get_start_end(h)
        # does not extend
        if h[end] <= original[end]:
            continue
        # overlaps
        if h[start] <= original[end]:
            # extend end ... already dealt with if new end is less
            original = (original[start], h[end])
        # there is a gap in between
        else:
            length += (original[end] - original[start]) + 1
            original = (h[start], h[end])

    length += (original[end] - original[start]) + 1 # deal with the last one
    return length


# flip orientation if nec.
def get_start_end(h):
    start = int(h[0])
    end = int(h[1])
    if start > end:
        return end, start

    return start, end


def orthomcl_blast_parser(blast_file, fasta_dir, db_dir, nm):

    # create connection to DB
    con = lite.connect(os.path.join(db_dir, "orthoDB.db"))
    with con:
        #global cur
        cur = con.cursor()

        prev_subjectid = ''
        prev_queryid = ''
        # hash to hold subject info
        subject = {}

        # Set progress information
        if nm:
            if nm.stop:
                raise KillByUser("")
            total = 0
            for total, _ in enumerate(open(blast_file)):
                pass
            nm.total = total
            nm.msg = None
            nm.counter = 0

        # parse fasta files
        genes = get_genes(fasta_dir)
        blast_fh = open(blast_file, "r")

        for line in blast_fh:

            if nm:
                if nm.stop:
                    raise KillByUser("")
                nm.counter += 1

            splitted = line.split()

            query_id = splitted[0]
            subject_id = splitted[1]
            percent_identity = splitted[2]
            length = int(splitted[3])
            query_start = splitted[6]
            query_end = splitted[7]
            subject_start = splitted[8]
            subject_end = splitted[9]
            evalue = splitted[10]

            if query_id != prev_queryid or subject_id != prev_subjectid:

                # print previous subject
                if subject:
                    print_previous_subject(subject, cur)

                # initialize new one from first HSP
                prev_subjectid = subject_id
                prev_queryid = query_id

                # from first hsp
                tup = format_evalue(evalue)

                subject = {"queryId": query_id}
                subject["subjectId"] = subject_id
                subject["queryShorter"] = get_taxon_and_length(subject, genes)

                subject["evalueMant"] = tup[0]
                subject["evalueExp"] = tup[1]
                subject["totalIdentities"] = 0
                subject["totalLength"] = 0
                subject["hspspans"] = []

            # get additional info from subsequent HSPs
            hspspan = (subject_start, subject_end)
            if subject and subject["queryShorter"]:
                hspspan = (query_start, query_end)
            subject["hspspans"].append(hspspan)
            subject["totalIdentities"] += float(percent_identity) * length
            subject["totalLength"] += length

        print_previous_subject(subject, cur)

    con.close()


# if __name__ == "__main__":
#     orthomcl_blast_parser("AllVsAll.out", "compliantFasta", ".")


__author__ = "Fernando Alves and Diogo N. Silva"



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import sqlite3 as lite
import os

try:
    from process.error_handling import KillByUser
except ImportError:
    from trifusion.process.error_handling import KillByUser


def printInparalogsFile (cur, filename, nm=None):

    cur.execute("select taxon_id, sequence_id_a, sequence_id_b, normalized_score\
        from InParalog\
        order by taxon_id, sequence_id_a, sequence_id_b asc")

    file_fh = open(filename, "w")

    with file_fh:
        while True:

            if nm:
                if nm.stop:
                    raise KillByUser("")

            row = cur.fetchone()
            if row is None:
                break

            file_fh.write("{}\t{}\t{}\n".format(row[1],
                                                row[2],
                                                str((float(row[3]) * 1000 + .5) / 1000)))


################################################################


def printOrthologsFile (cur, filename, nm=None):

    cur.execute("select taxon_id_a, taxon_id_b, sequence_id_a, sequence_id_b, normalized_score\
        from Ortholog\
        order by taxon_id_a, taxon_id_b, sequence_id_a, sequence_id_b asc")

    file_fh = open(filename, "w")

    with file_fh:
        while True:

            if nm:
                if nm.stop:
                    raise KillByUser("")

            row = cur.fetchone()
            if row is None:
                break

            file_fh.write("{}\t{}\t{}\n".format(row[2],
                                                row[3],
                                                str((float(row[4]) * 1000 + .5) / 1000)))

################################################################


def printCoOrthologsFile (cur, filename, nm=None):

    cur.execute("select taxon_id_a, taxon_id_b, sequence_id_a, sequence_id_b, normalized_score\
        from CoOrtholog\
        order by taxon_id_a, taxon_id_b, sequence_id_a, sequence_id_b asc")

    file_fh = open(filename, "w")

    with file_fh:
        while True:

            if nm:
                if nm.stop:
                    raise KillByUser("")

            row = cur.fetchone()
            if row is None:
                break

            file_fh.write("{}\t{}\t{}\n".format(row[2],
                                                row[3],
                                                str((float(row[4]) * 1000 + .5) / 1000)))

################################################################


def printMclAbcFile (cur, filename, nm=None):

    cur.execute("select sequence_id_a, sequence_id_b, normalized_score\
        from InParalog\
        union\
        select sequence_id_a, sequence_id_b, normalized_score\
        from Ortholog\
        union\
        select sequence_id_a, sequence_id_b, normalized_score\
        from CoOrtholog")

    file_fh = open(filename, "w")

    with file_fh:
        while True:

            if nm:
                if nm.stop:
                    raise KillByUser("")

            row = cur.fetchone()
            if row is None:
                break

            file_fh.write("{}\t{}\t{}\n".format(row[0],
                                                row[1],
                                                str((float(row[2]) * 1000 + .5) / 1000)))


def execute(db_dir, dest, nm=None):

    con = lite.connect(os.path.join(db_dir, "orthoDB.db"))

    # Set up progression information
    if nm:
        if nm.stop:
            raise KillByUser("")
        nm.total = 4
        nm.counter = 0

    with con:

        cur = con.cursor()

        if nm:
            if nm.stop:
                raise KillByUser("")
            nm.counter = 1

        printOrthologsFile(cur, os.path.join(dest, "backstage_files",
                                             "orthologs.txt"), nm=nm)
        if nm:
            if nm.stop:
                raise KillByUser("")
            nm.counter = 2

        printInparalogsFile(cur, os.path.join(dest, "backstage_files",
                                              "inparalogs.txt"), nm=nm)
        if nm:
            if nm.stop:
                raise KillByUser("")
            nm.counter = 3
        printCoOrthologsFile(cur, os.path.join(dest, "backstage_files",
                                               "coorthologs.txt"), nm=nm)
        if nm:
            if nm.stop:
                raise KillByUser("")
            nm.counter = 4
        printMclAbcFile(cur, os.path.join(dest, "backstage_files",
                                          "mclInput"), nm=nm)

    con.close()

if __name__ == "__main__":
    execute(".", ".")

__author__ = "Fernando Alves and Diogo N. Silva"



"""
The `base` sub-package contains three modules:

`html_creator`
--------------

A basic HTML template creator for Orthology's full report.

`plotter`
---------
Contains all functions used to generate plots for TriFusion's
Statistics screen and for TriStats.

`sanity`
--------
Contains several sanity checks performed during the execution
of the TriSeq CLI program.
"""



try:
    from setuptools import setup
except ImportError:
    import ez_setup
    ez_setup.use_setuptools()
    from setuptools import setup

import trifusion

VERSION = trifusion.__version__

with open('README.md') as f:
    readme = f.read()


def mcl_data_files():
    """
    Automatically detects the platform and sets the appropriate mcl binary
    """

    data_file = ["trifusion/data/resources/mcl/linux/mcl",
                 "trifusion/data/resources/mcl/MacOS/mcl",
                 "trifusion/data/resources/mcl/windows/64bit/mcl64.exe",
                 "trifusion/data/resources/mcl/windows/64bit/cygwin1.dll",
                 "trifusion/data/resources/mcl/windows/32bit/mcl32.exe",
                 "trifusion/data/resources/mcl/windows/32bit/cygwin1.dll"]

    return data_file


mcl_file = mcl_data_files()

setup(
    name="trifusion",
    version=VERSION,
    packages=["trifusion",
              "trifusion.base",
              "trifusion.data",
              "trifusion.data.backgrounds",
              "trifusion.data.resources",
              "trifusion.data.resources.theme",
              "trifusion.data.screens",
              "trifusion.ortho",
              "trifusion.process",
              "trifusion.progressbar"],
    package_data={"trifusion": ["*.kv"],
                  "trifusion.data.screens": ["*.kv"],
                  "trifusion.data.backgrounds": ["*.png", "*.ico"],
                  "trifusion.data.resources": [".desktop"]},
    install_requires=[
        "seaborn",
        "configparser",
        "pandas",
        "matplotlib",
        "numpy",
        "psutil",
        "scipy",
    ],
    description=("Streamlining phylogenomic data gathering, processing and "
                 "visualization"),
    long_description=readme,
    url="https://github.com/ODiogoSilva/TriFusion",
    author="Diogo N Silva",
    author_email="odiogosilva@gmail.com",
    license="GPL3",
    classifiers=["Development Status :: 4 - Beta",
                 "Intended Audience :: Science/Research",
                 "License :: OSI Approved :: GNU General Public License v3 ("
                 "GPLv3)",
                 "Natural Language :: English",
                 "Operating System :: POSIX :: Linux",
                 "Operating System :: MacOS :: MacOS X",
                 "Operating System :: Microsoft :: Windows",
                 "Programming Language :: Python",
                 "Programming Language :: Python :: 2.7",
                 "Topic :: Scientific/Engineering :: Bio-Informatics"],
    data_files=mcl_file,
    entry_points={
        "gui_scripts": [
            "TriFusion = trifusion.TriFusion:gui_exec"
        ],
        "console_scripts": [
            "orthomcl_pipeline = trifusion.orthomcl_pipeline:main",
            "TriSeq = trifusion.TriSeq:main",
            "TriStats = trifusion.TriStats:main"
        ]
    },
)


"""
How to use background tasks
===========================

Tasks that are too time consuming to be executed in TriFusion's main thread
are defined here. These functions should be called from
:class:`~trifusion.app.TriFusionApp` in a `threading.Thread` object. These
functions must rely on `multiprocessing.Namespace` and `Queue.Queue` objects
to transfer data between the worker and main thread.

Generic creation of background tasks in :class:`~trifusion.app.TriFusionApp`
----------------------------------------------------------------------------

Any method in :class:`~trifusion.app.TriFusionApp` can start a task in a
worker thread, provided it follows a few guidelines described below.
However, for most cases, there is already a
:meth:`~trifusion.app.TriFusionApp.run_in_background` method that greatly
facilitates this process::

    def run_in_background(self, func, second_func, args1, args2=None,
                          no_arg2=False, msg="Crunching data...",
                          cancel=True):

Suppose you have defined a function (`my_func`) in
:mod:`~trifusion.data.resources.background_tasks` that is meant to run in the
background. This function may take any number of arguments, but in this
example it takes only one. To execute this function in the background,
simply call :meth:`~trifusion.app.TriFusionApp.run_in_background` like this::

    run_in_background(my_func, None, args1=[my_arg])

This is the simplest example. In many cases `my_func` may return something,
and we may want to feed that something into another callback in the main
thread to update application structures or to perform any other task. This
can be easily accomplished with this method using the `second_func`
and `args2` arguments::

    run_in_background(my_func, my_callback, args1=[my_arg], args2=[other_arg])

In this case, the object returned by `my_func` will be passed directly
to `my_callback`. Since we also defined arguments to `my_callback`, the
argument list will be merged before calling `my_callback`. It's important
to note that `my_callback` is being called from the main thread, which
means that is can change application structures, but it can also freeze the
application window if it's too intensive.

Custom creation of background tasks
-----------------------------------

A :class:`~trifusion.app.TriFusionApp` method that need to executed some
functio in the background must follow some guidelines to ensure that
it will start and end properly.

Use a `Thread` object
~~~~~~~~~~~~~~~~~~~~~

A worker thread can be initiated like::

        p = threading.Thread(target=background_process,
                             args=(func, shared_ns, args1))
        p.start()

The background task is provided in the `target` argument, and any
potential arguments in the `args` argument.

Create and launch a waiting dialog
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

While the background task is being executed, a dialog of some sort should
be created in the main thread and ideally show some progress. Any
custom dialog can be created, but a general
:class:`~trifusion.data.resources.custom_widgets.CrunchData` dialog is already
available::

    content = CrunchData()
    # Add a custom message
    content.ids.msg.text = msg
    # Create popup with waiting dialog
    self.show_popup(title="", content=content, size=size,
                    separator_color=(0, 0, 0, 0),
                    border_color=tm.c_popup_border,
                    auto_dissmiss=False)

Schedule function that checks the worker thread's pulse
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

In order to check the pulse of the worker thread and/or receive information
from it while it's busy, a function can be scheduled to be called at
regular intervals using kivy's `Clock` object::

    # Schedule function that checks the process' pulse
    check_func = partial(check_process_status, p, shared_ns)
    Clock.schedule_interval(check_func, .1)

The `check_process_status` function may execute anything, like checking
the `Namespace` or `Queue` objects of the worker thread to update the
progress. The most important thing, however, is to check if the worker
thread is alive, and if not, unschedule itself, close the waiting popup,
join the thread, close any connections and relevant objects::

    def check_process_stats(self, p, shared_ns):

        if not p.is_alive():
            Clock.unschedule(check_func)
            self.dismiss_popup()
            p.join()

If the function in the worker thread returns some object, this woul be
the place to get that object and pass it to another callback::

    def check_process_stats(self, p, shared_ns):

        if not p.is_alive():

            obj = queue.get()
            self.my_callback(obj)

            Clock.unschedule(check_func)
            self.dismiss_popup()
            p.join()

Add a kill switch
~~~~~~~~~~~~~~~~~

Whenever possible, it's desirable to add a kill switch to the
`check_process_status` function, which changes the `Namespace.stop` attribute
to True, signaling the worker thread to stop::

    def check_process_stats(self, p, shared_ns):

        if self.terminate_background:
            shared_ns.stop = True
            time.sleep(1)
            self.dismiss_popup()
            Clock.unschedule(check_func)

"""

try:
    from process import data
    from process.error_handling import KillByUser,\
        MultipleSequenceTypes, EmptyAlignment
    from process.sequence import AlignmentList, Alignment
    import orthomcl_pipeline as ortho_pipe
    from ortho import OrthomclToolbox as OrthoTool
except ImportError:
    from trifusion.process import data
    from trifusion.process.error_handling import KillByUser,\
        MultipleSequenceTypes, EmptyAlignment
    from trifusion.process.sequence import AlignmentList, Alignment
    import trifusion.orthomcl_pipeline as ortho_pipe
    from trifusion.ortho import OrthomclToolbox as OrthoTool

from os.path import join, basename
from collections import OrderedDict
from copy import deepcopy
import logging
import shutil
import cPickle as pickle
import time

import os
os.environ["KIVY_NO_ARGS"] = "1"


def remove_tmp(temp_dir, sql_con):
    """Removes TriFusion's temporary directory and closes sqlite connection.

    Removes the temporary directory and all its contents and closes
    the connection to the sqlite database.

    Parameters
    ----------
    temp_dir : str
        Path to the temporary directory
    sql_con : sqlite3.Connection
        Sqlite3 connection object
    """

    # Give some time to child threads to exit
    time.sleep(1)

    # Close database connection
    sql_con.close()

    # Remove temporary files
    if os.path.exists(temp_dir):
        shutil.rmtree(temp_dir)

    return 1


def load_proc(aln_list, file_list, nm, queue):
    """Task that loads alignment files into TriFusion.

    Loads alignment files provided via the `file_list` argument into the
    `AlignmentList` object provided via `aln_list`.

    Parameters
    ----------
    aln_list : trifusion.process.sequence.AlignmentList
        AlignmentList object.
    file_list : list
        List of paths to alignment files.
    nm : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.
    queue :Queue.Queue
        Queue object used to transfer the AlignmentList to the main thread.
    """
    try:
        if aln_list:
            aln_list.resume_database()
            aln_list.add_alignment_files(file_list,
                                         ns=nm)
            aln_obj = aln_list
        else:
            aln_obj = AlignmentList(file_list, shared_namespace=nm)

        aln_obj.close_database()
        nm.res = aln_obj

    except MultipleSequenceTypes:
        nm.exception = "multiple_type"

    except IOError:
        return

    except KillByUser:
        return

    except Exception as e:
        logging.exception("Unexpected error when loading input data")
        print(e)


def get_stats_summary(dest, aln_list, active_file_set, active_taxa_set,
                        ns):
    """Calculates Statistic's summary statistics.

    Executes the `get_summary_stats` method in the background and writes
    the output in pickle files.

    Parameters
    ----------
    dest : str
        Path to directory where the pickle objects with the results will
        be created.
    aln_list : trifusion.process.sequence.AlignmentList
        AlignmentList object.
    active_file_set : list
        List with the active alignments via their `Alignment.name` attribute.
    active_taxa_set : list
        List with the active taxa.
    ns : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.
    """

    time.sleep(.4)

    try:
        # Creating deepcopy to perform changes  without impacting main
        # attribute
        main_aln = deepcopy(aln_list)
        main_aln.set_database_connections(aln_list.cur, aln_list.con)
        # Update alignment object according to active file and taxa sets
        main_aln.update_active_alignments(active_file_set)
        main_aln.update_taxa_names(active_taxa_set)

        with open(join(dest, "stats.pc"), "wb") as fh_stats, \
                open(join(dest, "table.pc"), "wb") as fh_table:

            # Check if active data sets are not empty. If so, raise an
            # exception
            if main_aln.alignments == OrderedDict() or not main_aln.taxa_names:
                for fh in [fh_stats, fh_table]:
                    pickle.dump({"exception": "Alignment is empty after file "
                                              "and taxa filters"}, fh)
                return

            stats = main_aln.get_summary_stats(ns=ns)
            table = main_aln.get_gene_table_stats()
            pickle.dump(stats, fh_stats)
            pickle.dump(table, fh_table)

    except KillByUser:
        return


def background_process(f, ns, a):
    """General execution of a background process.

    Allows a generic function to be executed in the background with
    or without arguments, provided via the `a` argument.

    Parameters
    ----------
    f : function
        Callback function.
    ns : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.
    a : list
        List of arguments provided to the `f` function. Can be None.
    """
    try:
        if a:
            if "use_ns" in a:
                a.remove("use_ns")
                val = f(*a, ns=ns)
            else:
                val = f(*a)
        else:
            val = f()
        ns.val = val

    except KillByUser:
        return

    except IOError as e:
        print(e)

    except Exception:
        logging.exception("Unexpected exit in %s", f.__name__)
        ns.exception = True


def background_export_groups(f, nm, a):
    """Specific callback for exporting Orthology groups.

    Parameters
    ----------
    f : function
        Callback function.
    nm : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.
    a : list
        List of arguments provided to the `f` function.
    """
    try:
        f(*a, shared_namespace=nm)

    except KillByUser:
        pass

    except IOError as e:
        print(e)

    except Exception as e:
        logging.exception("Unexpected error when exporting ortholog "
                          "groups")
        nm.exception = True


def orto_execution(nm, temp_dir, proteome_files, protein_min_len,
                   protein_max_stop, usearch_file, usearch_evalue,
                   usearch_threads, usearch_output, mcl_file, mcl_inflation,
                   ortholog_prefix, group_prefix, orto_max_gene,
                   orto_min_sp, sqldb, ortho_dir, usearch_db):
    """Execution of the orthology search pipeline.


    Executes all pipeline subprocesses sequentially and updates the
    Progess dialog label.

    Parameters
    ----------
    nm : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.

    temp_dir : str
        Path to TriFusion's temporary directory.
    proteome_files : list
        List of pahts to proteome files.
    protein_min_len : int
        Minimum lenght of protein sequences.
    protein_max_stop : int
        Maximum percentage of stop codons allowed.
    usearch_file : str
        Path to usearch executbale.
    usearch_evalue: int or float
        Evalue for usearch execution.
    usearch_threads : int
        Number of threads used by usearch execution.
    usearch_output : str
        Name of usearch's output file.
    mcl_file : str
        Path to the mcl executable.
    mcl_inflation : list
        List of inflation values (int) to perform at the end of the orthology
        search.
    ortholog_prefix : str
        Prefix for the name of the orthologs.
    group_prefix : str
        Prefix for the name of the group files.
    orto_max_gene : int
        Maximum number of gene copies allowed when filtering the search
        results.
    orto_min_sp : int
        Minimum number of taxa representation when filtering the search
        results.
    sqldb : str
        Path to the sqlite database.
    ortho_dir : str
        Path to the directory where the results will be generated.
    usearch_db : str
        Name of the file used as database for usearch.
    """

    try:
        nm.finished_tasks = []

        nm.task = "schema"
        ortho_pipe.install_schema(temp_dir)
        nm.finished_tasks = ["schema"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "adjust"
        ortho_pipe.adjust_fasta(proteome_files, ortho_dir, nm)
        nm.finished_tasks = ["schema", "adjust"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "filter"
        ortho_pipe.filter_fasta(protein_min_len, protein_max_stop,
                                usearch_db, ortho_dir, nm)
        nm.finished_tasks = ["schema", "adjust", "filter"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "usearch"
        ortho_pipe.allvsall_usearch(usearch_db, usearch_evalue, ortho_dir,
                                    usearch_threads, usearch_output,
                                    usearch_bin=usearch_file, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "parse"
        ortho_pipe.blast_parser(usearch_output, ortho_dir,
                                db_dir=temp_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "pairs"
        ortho_pipe.pairs(temp_dir, nm=nm)
        ortho_pipe.dump_pairs(temp_dir, ortho_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "mcl"
        ortho_pipe.mcl(mcl_inflation, ortho_dir, mcl_file=mcl_file, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs", "mcl"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "dump"
        ortho_pipe.mcl_groups(mcl_inflation, ortholog_prefix, "1000",
                              group_prefix, ortho_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs", "mcl", "dump"]

        if nm.stop:
            raise KillByUser("")

        nm.task = "filter_groups"
        stats, groups_obj = ortho_pipe.export_filtered_groups(
            mcl_inflation,
            group_prefix,
            orto_max_gene,
            orto_min_sp,
            sqldb + "_out",
            join(ortho_dir, "backstage_files", usearch_db),
            temp_dir,
            ortho_dir, nm=nm)
        nm.finished_tasks = ["schema", "adjust", "filter", "usearch", "parse",
                             "pairs", "mcl", "dump", "filter_groups"]

        if nm.stop:
            raise KillByUser("")

        # stats is a dictionary containing the inflation value as
        #  key and a list with the orthologs as value
        nm.stats = stats
        nm.groups = groups_obj

    except KillByUser:
        return

    except IOError as e:
        nm.exception = str(e)
        print(e)
        return

    except Exception as e:
        logging.exception("Unexpected exit in Orthology search")
        nm.exception = str(e)


def update_active_fileset(aln_obj, set_name, file_list, active_file_list,
                          file_groups, filename_map):
    """Upates the active files of an `AlignmentList` object

    This method is similar in purpose to
    `AlignmentList.update_active_alignments` but it can convert the set
    name of the active group defined in TriFusion to an actual list
    of files.

    Parameters
    ----------
    aln_obj : trifusion.process.sequence.AlignmentList
        AlignmentList object.
    set_name : str
        Name of the active file group.
    file_list : list
        List of alignment files loaded into TriFusion.
    file_groups : dict
        Maps the name of custom file groups to a list of alignment files.
    filename_map : dict
        Maps the basename of aligment files to their full path.
    """
    # Determine the selected active taxa set from the dropdown menu
    if set_name == "All files":
        aln_obj.update_active_alignments([x for x in file_list])
        return aln_obj
    if set_name == "Active files":
        aln_obj.update_active_alignments(active_file_list)
        return aln_obj
    else:
        aln_obj.update_active_alignments(
            [filename_map[x] for x in file_groups[set_name]])
        return aln_obj


def update_active_taxaset(aln_obj, set_name, active_taxa_list, taxa_groups):
    """Upates the active taxa of an `AlignmentList` object

    This method is similar in purpose to
    `AlignmentList.update_taxa_names` but it can convert the set
    name of the active group defined in TriFusion to an actual list
    of taxa.

    Parameters
    ----------
    aln_obj : trifusion.process.sequence.AlignmentList
        AlignmentList object.
    set_name : str
        Name of the active taxa group.
    active_taxa_list : list
        List of active taxa.
    taxa_groups : dict
        Maps the name of custom taxa groups to a list of taxon names.
    """

    if set_name == "All taxa":
        return aln_obj
    if set_name == "Active taxa":
        tx_set = active_taxa_list
    else:
        tx_set = taxa_groups[set_name]

    # Update active taxa
    aln_obj.update_taxa_names(tx_set)
    return aln_obj


def process_execution(aln_list, file_set_name, file_list, active_file_list,
                        file_groups, filename_map,
                        taxa_set_name, active_taxa_list, ns, taxa_groups,
                        hap_prefix, secondary_operations, secondary_options,
                        missing_filter_settings, taxa_filter_settings,
                        codon_filter_settings, variation_filter_settings,
                        output_file, rev_infile, main_operations, zorro_suffix,
                        partitions_file, output_formats, create_partfile,
                        use_nexus_partitions, use_nexus_models,
                        phylip_truncate_name, output_dir, use_app_partitions,
                        consensus_type, ld_hat, ima2_params,
                        conversion_suffix, state_fl):
    """The Process execution

    Parameters
    ----------
    aln_list : trifusion.process.sequence.AlignmentList
        AlignmentList object.
    file_set_name : str
        Name of the active file group.
    file_list : list
        List of alignment files loaded into TriFusion.
    file_groups : dict
        Maps the name of custom file groups to a list of alignment files.
    filename_map : dict
        Maps the basename of aligment files to their full path.
    taxa_set_name : str
        Name of the active taxa group.
    active_taxa_list : list
        List of active taxa.
    ns : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.
    taxa_groups : dict
        Maps the name of custom taxa groups to a list of taxon names.
    hap_prefix : str
        See :attr:`~trifusion.app.TriFusionApp.hap_prefix` attribute.
    secondary_operations : dict
        See :attr:`~trifusion.app.TriFusionApp.secondary_operations` attribute.
    secondary_options : dict
        See :attr:`~trifusion.app.TriFusionApp.secondary_options` attribute.
    missing_filter_settings : list
        See :attr:`~trifusion.app.TriFusionApp.missing_filter_settings`
        attribute.
    taxa_filter_settings : list
        See :attr:`~trifusion.app.TriFusionApp.taxa_filter_settings` attribute.
    codon_filter_settings: list
         See :attr:`~trifusion.app.TriFusionApp.codon_filter_settings`
         attribute.
    variation_filter_settings : list
        See :attr:`~trifusion.app.TriFusionApp.variation_filter_settings`
        attribute.
    output_file : str
        Name of the output file.
    rev_infile : str
        See :attr:`~trifusion.app.TriFusionApp.rev_infile` attribute.
    main_operations : dict
        See :attr:`~trifusion.app.TriFusionApp.main_operations` attribute.
    zorro_suffix : str
        See :attr:`~trifusion.app.TriFusionApp.zorro_suffix` attribute.
    partitions_file : str
        See :attr:`~trifusion.app.TriFusionApp.partitions_file` attribute.
    output_formats : list
        See :attr:`~trifusion.app.TriFusionApp.output_formats` attribute.
    create_partfile : bool
        See :attr:`~trifusion.app.TriFusionApp.create_partfile` attribute.
    use_nexus_partitions : bool
        See :attr:`~trifusion.app.TriFusionApp.use_nexus_partitions` attribute.
    use_nexus_models : bool
        See :attr:`~trifusion.app.TriFusionApp.use_nexus_models` attribute.
    phylip_truncate_name : bool
        See :attr:`~trifusion.app.TriFusionApp.phylip_truncate_name`
        attribute.
    output_dir : str
        Path to directory where the output file(s) will be generated.
    use_app_partitions : bool
        See :attr:`~trifusion.app.TriFusionApp.use_app_partitions` attribute.
    consensus_type : str
        Mode of consensus variation handling.
    ld_hat : bool
        See :attr:`~trifusion.app.TriFusionApp.ld_hat` attribute.
    ima2_params :
        See :attr:`~trifusion.app.TriFusionApp.ima2_options` attribute.
    conversion_suffix : str
        See :attr:`~trifusion.app.TriFusionApp.conversion_suffix` attribute.

    """

    # Give just a little time for the waiting dialog to open without lag
    time.sleep(.1)

    def reverse_concatenation(aln, table_in=None, table_out=None):
        """ Wrapper of the reverse concatenation operation

        Parameters
        ----------
        aln : trifusion.process.sequence.AlignmentList
        AlignmentList object.
        """

        con = aln.con

        if use_app_partitions:

            # If there are multiple alignments, concatenate them first and
            # then apply the reverse concatenation according to the partitions
            # defined in the app
            if len(aln.alignments) > 1:
                aln.concatenate(table_in=table_in, table_out=table_out)

            aln.reverse_concatenate(table_in=table_in, table_out=table_out,
                                    ns=ns)

        else:

            # Remove all alignments except the one to be reversed
            al = [x for x in aln.all_alignments if x != rev_infile]
            aln.remove_file(al)

            er = aln.partitions.read_from_file(partitions_file)

            if er:
                ns.exception = {
                    "exception": ["Invalid partition file", er.message]}
                raise data.InvalidPartitionFile("")

            aln.reverse_concatenate(table_in=table_in, table_out=table_out,
                                    ns=ns)

    def filter_aln(aln, table_in, table_out="_filter"):
        """Wrapper for alignment filtering operations

        Parameters
        ----------
        aln : trifusion.process.sequence.AlignmentList
            AlignmentList object.
        table_out : str
            Specifies the output table for filtering methods.
        """

        # Check if a minimum taxa representation was specified
        if secondary_options["gap_filter"]:
            if missing_filter_settings[1][0]:
                ns.main_msg = "Filter (minimum taxa)"
                aln.filter_min_taxa(missing_filter_settings[1][1], ns=ns)

        # Filter by taxa
        if secondary_options["taxa_filter"]:
            # Get taxa list from taxa groups
            ns.main_msg = "Filter (by taxa)"
            taxa_list = taxa_groups[taxa_filter_settings[1]]
            aln.filter_by_taxa(taxa_list, taxa_filter_settings[0], ns=ns)

        # Filter codon positions
        if secondary_options["codon_filter"]:
            ns.main_msg = "Filter (by codon)"
            aln.filter_codon_positions(codon_filter_settings,
                                       table_in=table_in,
                                       table_out=table_out, ns=ns)

        # Filter missing data
        if secondary_options["gap_filter"]:
            ns.main_msg = "Filter (by missing data)"
            if missing_filter_settings[0][0]:
                aln.filter_missing_data(missing_filter_settings[0][1],
                                        missing_filter_settings[0][2],
                                        table_in=table_in,
                                        table_out=table_out, ns=ns)

        # Filter variation
        if secondary_options["variation_filter"]:
            # Checks for variable site filter
            if variation_filter_settings[0] or variation_filter_settings[1]:
                ns.main_msg = "Filter (by variable sites)"
                aln.filter_segregating_sites(variation_filter_settings[0],
                                             variation_filter_settings[1],
                                             table_in=table_out, ns=ns)
            # Checks for informative site filter
            if variation_filter_settings[2] or variation_filter_settings[3]:
                ns.main_msg = "Filter (by informative sites)"
                aln.filter_informative_sites(variation_filter_settings[2],
                                             variation_filter_settings[3],
                                             table_in=table_out, ns=ns)

        # Pipe the information on the filtered alignments to the main process
        # only if it was applied a filter that changes the final alignments
        if set(aln.filtered_alignments.values()) != {None}:
            ns.filtered_alns = aln.filtered_alignments

        # Reset main label text
        ns.main_msg = None

        # Some filter configurations may result in empty final alignment
        # list. In such cases, return and issue warning
        if not aln.alignments:
            ns.exception = {
                "exception": ["Empty alignment",
                              "The alignment is empty after applying filters"]}
            raise EmptyAlignment("Active alignment is empty")

        return aln

    def concatenation(aln, table_in, table_out):
        """Wrapper for concatenation operation

        Parameters
        ----------
        aln : trifusion.process.sequence.AlignmentList
            AlignmentList object.
        table_in : str
            Specifies the input table for concatenation.
        """

        if secondary_options["zorro"]:
            ns.msg = "Concatenating ZORRO files"
            zorro_data = data.Zorro(aln, zorro_suffix)
            zorro_data.write_to_file(output_file)

        aln.concatenate(table_in=table_in, table_out=table_out, ns=ns)

        # Sets the single alignment to True, for other method to be aware of
        # this
        ns.sa = True

        return aln

    def consensus(aln, table_in, table_out):
        """Wrapper of the consensus operation

        Parameters
        ----------
        aln : trifusion.process.sequence.AlignmentList
            AlignmentList object.
        table_in : str
            Specifies the input data table.
        table_out : str
            Specifies the output table for filtering methods.
        """

        aln.consensus(consensus_type=consensus_type,
                      single_file=secondary_options["consensus_single"],
                      table_in=table_in,
                      table_out=table_out,
                      ns=ns)

    def writer(aln, filename=None, suffix_str="", conv_suffix="",
               table_name=None):
        """Wrapper for the output writing operations

        Parameters
        ----------
        aln : trifusion.process.sequence.AlignmentList
            AlignmentList object.
        filename : str
            If provided, will overwrite the `output_file` variable.
        suffix_str : str
            Provides the suffix for the `AlignmentList.write_to_file`
            method argument.
        conv_suffix : str
            Provides the suffix provided for
            the conversion of files. This suffix will always precede the
            suffix_str, which is meant to apply suffixes specific to secondary
            operations.
        table_suffix : str
            Suffix of the table from where the sequence data will be
            retrieved.
        table_name : str
            Name of the table from where the sequence data will be retrieved.
        """

        try:
            if filename:
                outfile = filename
            else:
                outfile = output_file

            # The output file(s) will only be written after all the required
            # operations have been concluded. The reason why there are two "if"
            # statement for "concatenation" is that the input alignments must
            # be concatenated before any other additional operations. If the
            # first if statement did not exist, then all additional options
            # would have to be manually written for both "conversion" and
            # "concatenation". As it is, when "concatenation", the aln_obj is
            # firstly converted into the concatenated alignment, and then all
            # additional operations are conducted in the same aln_obj

            aln.write_to_file(
                output_formats,
                output_file=outfile,
                output_suffix=suffix_str,
                conversion_suffix=conv_suffix,
                interleave=secondary_options["interleave"],
                partition_file=create_partfile,
                output_dir=output_dir,
                use_charset=use_nexus_partitions,
                phy_truncate_names=phylip_truncate_name,
                ld_hat=ld_hat,
                ima2_params=ima2_params,
                use_nexus_models=use_nexus_models,
                ns_pipe=ns,
                table_name=table_name,
                upper_case=secondary_options["upper_case"])

        except IOError as e:
            logging.exception(e)

    try:

        # If concatenation was no specified, reset the output_file variable
        if not main_operations["concatenation"]:
            output_file = None

        aln_list.save_state(state_fl)

        # aln_object = deepcopy(aln_list)
        # Restore database connections, since they are broken during the
        # deepcopy operation
        # aln_object.set_database_connections(aln_list.cur, aln_list.con)

        # Setting the alignment to use.
        # Update active file set of the alignment object
        aln_object = update_active_fileset(aln_list,
                                           file_set_name,
                                           file_list,
                                           active_file_list,
                                           file_groups,
                                           filename_map)

        # Update active taxa set of the alignment object
        main_aln = update_active_taxaset(aln_object, taxa_set_name,
                                           active_taxa_list,
                                           taxa_groups)

        ns.proc_files = len(aln_object.alignments)

        # Initialize attribute tha will store the number of filtered
        # alignments for reporting purposes
        ns.filtered_alns = None

        # The execution of the process module will begin with all the
        # operations on the main output alignment. Only after the main
        # output file has been created will the additional secondary output
        # files be processed.

        #####
        # Perform operations on MAIN OUTPUT
        #####

        # Set the name of the sqlite table harboring the main output
        # alignment if any of the secondary operations is specified
        main_table = "main_output"

        # Reverse concatenation
        # Active table: Based on partition names
        if main_operations["reverse_concatenation"]:
            ns.task = "reverse_concatenation"
            reverse_concatenation(main_aln, table_in=main_table,
                                  table_out=main_table)
            ns.finished_tasks.append("reverse_concatenation")

        # Filtering
        # Active table: * / *main
        if secondary_options["collapse_filter"] and not \
                secondary_options["collapse_file"]:
            ns.task = "collapse"
            # If the the collapse filter is active, perform this
            # filtering first. This is because the filter will allow 0% of
            # missing data, which will always be as stringent or more than any
            # missing data filter set.
            main_aln.filter_missing_data(0, 0,table_in=main_table,
                                         table_out=main_table, ns=ns)
            ns.finished_tasks.append("collapse_filter")

        # Active table: * / *main
        if secondary_operations["filter"] and not \
                secondary_options["filter_file"]:
            ns.task = "filter"
            main_aln = filter_aln(main_aln, table_in=main_table,
                                  table_out=main_table)
            ns.finished_tasks.append("filter")
        # Concatenation
        # Active table: concatenation
        if main_operations["concatenation"]:
            ns.task = "concatenation"
            main_aln = concatenation(main_aln, table_in=main_table,
                                     table_out=main_table)
            ns.finished_tasks.append("concatenation")
        # Collapsing
        # Active table: *main / concatenationmain
        if secondary_operations["collapse"] and not \
                secondary_options["collapse_file"]:
            ns.task = "collapse"
            hap_file = output_file if output_file else None
            main_aln.collapse(haplotype_name=hap_prefix, dest=output_dir,
                              table_out=main_table, ns=ns, table_in=main_table,
                              haplotypes_file=hap_file)
            ns.finished_tasks.append("collapse")
        # Gcoder
        # Active table: *main / concatenationmain
        if secondary_operations["gcoder"] and not \
                secondary_options["gcoder_file"]:
            ns.task = "gcoder"
            main_aln.code_gaps(table_in=main_table, table_out=main_table,
                               ns=ns)
            ns.finished_tasks.append("gcoder")
        # Consensus
        # Active table: *main / concatenationmain / consensus
        if secondary_operations["consensus"] and not \
                secondary_options["consensus_file"]:
            ns.task = "consensus"
            consensus(main_aln, table_in=main_table, table_out=main_table)
            ns.finished_tasks.append("consensus")

        ns.task = "write"
        writer(main_aln, conv_suffix=conversion_suffix,
               table_name=main_table)
        ns.finished_tasks.append("write")

        if not any(secondary_operations.values()):
            return

        #####
        # Perform operations on ADDITIONAL OUTPUTS
        #####

        # Table names for each secondary operation
        op_tables = {
            "filter": "filtered",
            "consensus": "consensus",
            "gcoder": "coded",
            "collapse": "collapsed"
        }

        # Stores operations that must be performed before concatenation,
        # if it was specified
        before_conc = ["filter"]

        # The additional outputs workflow has two basic flows. In the first,
        # it performs tasks that must be executed before concatenating.
        # In the second, performs tasks that are executed after concatenating.
        # In this last case, the concatenated data can be generated only once.

        # Perform the filtering and consensus option separately, since these
        # must be done before concatenation
        for op in [x for x, y in secondary_operations.items() if
                   x in before_conc and y and
                   secondary_options["%s_file" % x]]:

            if op == "filter":

                ns.task = "filter"

                # Remove previous temporary tables
                aln_object.remove_tables()

                # main_aln = deepcopy(aln_list)
                aln_list.restore_state(state_fl)
                # main_aln.set_database_connections(aln_list.cur, aln_list.con)

                ns.msg = "Creating additional filtered alignments(s)"
                suffix = "_filtered"
                table_name = op_tables[op]
                main_aln = filter_aln(aln_list, table_in=table_name,
                                      table_out=table_name)

            if main_operations["concatenation"]:

                filename = output_file + "_{}".format(table_name)

                ns.main_msg = "Concatenating"
                main_aln = concatenation(main_aln, table_in=table_name,
                                         table_out=table_name)
                ns.main_msg = "Writing output"
                writer(main_aln, filename=filename, table_name=table_name)
            else:
                ns.main_msg = "Writing output"
                writer(main_aln, suffix_str="_{}".format(table_name),
                       table_name=table_name, conv_suffix=conversion_suffix)

        # Remove previous temporary tables
        aln_object.remove_tables()

        # In the second part, we execute the tasks that can be performed after
        # the concatenation.

        # main_aln = deepcopy(aln_list)
        # main_aln.set_database_connections(aln_list.cur, aln_list.con)
        aln_list.restore_state(state_fl)

        concatenated = False

        ns.sa = False

        for op in [x for x, y in secondary_operations.items() if
                   x not in before_conc and y and
                   secondary_options["%s_file" % x]]:

            ns.task = op

            table_name = op_tables[op]

            if main_operations["concatenation"]:
                if op == "collapse" and secondary_options["collapse_filter"]:
                    pass
                else:
                    if not concatenated:
                        aln_list = concatenation(aln_list,
                                                 table_in="add_concat",
                                                 table_out="add_concat")
                        aln_list.master_table = "add_concat"
                        concatenated = True

            if op == "consensus":

                suffix = "_consensus"
                if main_operations["concatenation"]:
                    aln_list.alignment_idx.values()[0].name = op
                consensus(aln_list, table_in=table_name, table_out=table_name)

            elif op == "gcoder":
                suffix = "_coded"
                aln_list.code_gaps(table_in=table_name, table_out=table_name,
                                   ns=ns)

            elif op == "collapse":

                suffix = "_collapsed"

                # In this case, filter the unconcatenated alignment
                # and concatenate again
                if secondary_options["collapse_filter"]:
                    ns.main_msg = "Filtering for collapse"
                    aln_list.remove_tables()
                    aln_list.filter_missing_data(0, 0, ns=ns,
                                                   table_in=table_name,
                                                   table_out=table_name)
                    if main_operations["concatenation"]:
                        ns.main_msg = "Concatenation"
                        aln_list.concatenate(ns=ns, table_in=table_name,
                                               table_out=table_name)

                ns.main_msg = "Collapse"
                hap_file = output_file if output_file else None
                aln_list.collapse(haplotype_name=hap_prefix,
                                  conversion_suffix=conversion_suffix,
                                  haplotypes_file=hap_file,
                                  dest=output_dir, table_in=table_name,
                                  table_out=table_name, ns=ns)

            ns.main_msg = "Writing output"
            if main_operations["concatenation"]:
                filename = output_file + suffix
                writer(aln_list, filename=filename, table_name=table_name)
            else:
                writer(aln_list, suffix_str=suffix,
                       conv_suffix=conversion_suffix, table_name=table_name)

        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)

    except KillByUser:
        pass

    except IOError as e:
        logging.exception(e)
        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)
        return

    except EmptyAlignment:
        logging.exception("Empty alignment")
        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)
        # ns.exception = "EmptyAlignment"

    except Exception as e:
        print(e)
        # Log traceback in case any unexpected error occurs. See
        # self.log_file for whereabouts of the traceback
        logging.exception("Unexpected exit in Process execution")
        # Resets the taxa_names attribute of the aln_obj to include all taxa
        # aln_object.update_taxa_names(all_taxa=True)
        if not hasattr(ns, "exception"):
            ns.exception = {
                "exception": ["Unknown",
                              "Unexpected error when generating Process "
                              "output. Check the app logs."]}


def load_group_files(group_files, temp_dir, ns=None):
    """Task that loads orthology group files into TriFusion

    Parameters
    ----------
    group_files : list
        List of paths to group files.
    temp_dir : str
        Temporary directory where sqlite database will be created.
    ns : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.

    Returns
    -------
    og : trifusion.ortho.OrthomclToolbox.MultiGroupsLight
        `MultiGroupsList` object.
    og.filters : list
        List of filters for the `MultiGroupsList` object.
    """
    og = OrthoTool.MultiGroupsLight(db_path=temp_dir,
                                    groups=group_files,
                                    ns=ns)
    return [og, og.filters]


def orto_update_filters(ortho_groups, gn_filter, sp_filter, excluded_taxa,
                        group_names=None, default=False):
    """Task that updates filters of a `MultiGroupsLight` object

    Parameters
    ----------
    ortho_groups : trifusion.ortho.OrthomclToolbox.MultiGroupsLight
        `MultiGroupsList` object.
    gn_filter : int
        Filter for maximum gene copies.
    sp_filter : int
        Filter for minimum taxa representation.
    excluded_taxa : list
        List of taxa to be excluded.
    group_names : list
        List with name of group files.
    default : bool
        If True, the default filters will be used.

    Returns
    -------
    orto_groups :  trifusion.ortho.OrthomclToolbox.MultiGroupsLight
        `MultiGroupsList` object.
    """
    try:
        if group_names or group_names == []:
            ortho_groups.update_filters(gn_filter, sp_filter, excluded_taxa,
                                        group_names, default=default)
        else:
            ortho_groups.update_filters(gn_filter, sp_filter, excluded_taxa,
                                        default=default)
    except EOFError:
        pass
    return [ortho_groups]


def get_active_group(ortho_groups, old_active_group, active_group_name):
    """Task that retrieves the active `GroupLight` object.

    Parameters
    ----------
    ortho_groups : trifusion.ortho.OrthomclToolbox.MultiGroupsLight
        `MultiGroupsList` object.
    old_active_group : str
        Previous active `GroupLight` object.
    active_group_name :
        Name of the `GroupLight` object that will be active.

    Returns
    -------
    active_group : trifusion.ortho.OrthomclToolbox.GroupLight
        `GroupLight` object.
    """

    if not old_active_group:
        active_group = ortho_groups.get_group(active_group_name)
    else:
        if active_group_name == old_active_group.name:
            return [None]
        else:
            active_group = ortho_groups.get_group(active_group_name)

    return [active_group]


def get_orto_data(active_group, plt_idx, filt, exclude_taxa):
    """Creates plot data for orthology

    Given a GroupLight object, this function will execute the method that
    corresponds to plt_idx to generate data.

    Parameters
    ----------
    active_group : trifusion.ortho.OrthomclToolbox.GroupLight
        `GroupLight` object.
    plt_idx : str
        Identifier of the plot type that must have a correspondence in
        the method dictionary below.
    filt : list
        List with orthology filters.
    exclude_ taxa : list
        List of taxa to be excluded.
    """

    # Store the plot generation method in a dictionary where keys are
    # the text attributes of the plot spinner and the values are
    # bound methods
    methods = {
        "Taxa distribution": active_group.bar_species_distribution,
        "Taxa coverage": active_group.bar_species_coverage,
        "Gene copy distribution": active_group.bar_genecopy_distribution,
        "Taxa gene copies": active_group.bar_genecopy_per_species
    }

    # Check for excluded taxa. If any have been provided and are different from
    # the ones already set in the GroupLight object, then update the taxa
    # filter.
    if (exclude_taxa and exclude_taxa != active_group.excluded_taxa) or \
            (exclude_taxa == [] and
             exclude_taxa != active_group.excluded_taxa):
        # The update_stats flag of the exclude_taxa method is set to True
        # to update the group summary stats. This is important for eventual
        # corrections to the ortholog cluster filters.
        active_group.exclude_taxa(exclude_taxa, True)

    if filt:
        # If filters AND excluded taxa have been provided, the first thing
        # to do is check whether the provided filters are still correct.
        # Excluding taxa may lead to changes in the maximum values of the
        # ortholog filters, and this needs to be corrected here
        if exclude_taxa and exclude_taxa != active_group.excluded_taxa:
            # Correct gene copy filter maximum
            gn_filt = filt[0] if filt[0] <= active_group.max_extra_copy \
                else active_group.max_extra_copy
            # Correct minimum taxa representation maximum
            sp_filt = filt[1] if \
                filt[1] <= len(active_group.species_list) \
                else len(active_group.species_list)
        else:
            # No taxa have been excluded this time, so we'll keep the provided
            # filters
            gn_filt, sp_filt = filt

        active_group.update_filters(gn_filt, sp_filt, True)

    plot_data = methods[plt_idx](filt=filt)

    return [plot_data]


def get_stats_data(aln_obj, stats_idx, active_file_set, active_taxa_set,
                     additional_args, ns=None):
    """Task that retrieves the plot data from `AlignmentList` plot methods

    Given an aln_obj, this function will execute the according method to
    generate plot data.

    Parameters
    ----------
    aln_obj : `trifusion.process.sequence.AlignmentList`
        AlignmentList object.
    stats_idx : str
        Identifier of the method type that must have a correspondence in
        the `methods` dictionary below.
    active_filte_set : str
        Name of the active file set.
    active_taxa_set : str
        Name of the active taxa set.
    additional_args : dict
        Dictionary with keyword arguments that can be provided when the
        plot data method is called.
    ns : multiprocessing.Namespace
        Namespace object that allows communication between main and worker
        threads.
    """

    # Creating deepcopy to perform changes without impacting the main attribute
    main_aln = deepcopy(aln_obj)
    main_aln.set_database_connections(aln_obj.cur, aln_obj.con)

    # Update alignment object according to active file and taxa sets
    main_aln.update_active_alignments(active_file_set)
    main_aln.update_taxa_names(active_taxa_set)

    # Check if active data sets are not empty. If so, raise an exception
    if main_aln.alignments == OrderedDict() or not main_aln.taxa_names:
        return [EmptyAlignment("Active alignment is empty")]

    # List of gene specific idx. These plots only have one gene for the footer
    gene_specific = ["Pairwise sequence similarity gn"]

    if stats_idx in gene_specific:
        footer = [1, len(main_aln.taxa_names)]
    else:
        footer = [len(main_aln.alignments), len(main_aln.taxa_names)]

    methods = {"Gene occupancy": main_aln.gene_occupancy,
               "Distribution of missing data sp":
                   main_aln.missing_data_per_species,
               "Distribution of missing data":
                   main_aln.missing_data_distribution,
               "Distribution of missing orthologs":
                   main_aln.missing_genes_per_species,
               "Distribution of missing orthologs avg":
                   main_aln.missing_genes_average,
               "Distribution of sequence size":
                   main_aln.average_seqsize_per_species,
               "Distribution of sequence size all": main_aln.average_seqsize,
               "Cumulative distribution of missing genes":
                   main_aln.cumulative_missing_genes,
               "Sequence conservation gn":
                   main_aln.sequence_conservation_gnp,
               "Proportion of nucleotides or residues gn":
                   main_aln.characters_proportion_gene,
               "Proportion of nucleotides or residues gn prop":
                   main_aln.characters_proportion_gene,
               "Proportion of nucleotides or residues":
                   main_aln.characters_proportion,
               "Proportion of nucleotides or residues sp":
                   main_aln.characters_proportion_per_species,
               "Pairwise sequence similarity": main_aln.sequence_similarity,
               "Pairwise sequence similarity sp":
                   main_aln.sequence_similarity_per_species,
               "Pairwise sequence similarity gn":
                   main_aln.sequence_similarity_gene,
               "Segregating sites": main_aln.sequence_segregation,
               "Segregating sites sp": main_aln.sequence_segregation_per_species,
               "Segregating sites gn": main_aln.sequence_segregation_gene,
               "Segregating sites prop": main_aln.sequence_segregation,
               "Alignment length/Polymorphism correlation":
                   main_aln.length_polymorphism_correlation,
               "Distribution of taxa frequency":
                   main_aln.taxa_distribution,
               "Allele Frequency Spectrum":
                   main_aln.allele_frequency_spectrum,
               "Allele Frequency Spectrum prop":
                   main_aln.allele_frequency_spectrum,
               "Allele Frequency Spectrum gn":
                   main_aln.allele_frequency_spectrum_gene,
               "Missing data outliers": main_aln.outlier_missing_data,
               "Missing data outliers sp": main_aln.outlier_missing_data_sp,
               "Segregating sites outliers": main_aln.outlier_segregating,
               "Segregating sites outliers sp": main_aln.outlier_segregating_sp,
               "Sequence size outliers sp": main_aln.outlier_sequence_size_sp,
               "Sequence size outliers": main_aln.outlier_sequence_size}

    if additional_args:
        plot_data = methods[stats_idx](ns=ns, **additional_args)
    else:
        plot_data = methods[stats_idx](ns)

    return [plot_data, footer]


__author__ = "Diogo N. Silva"



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
#
#  Copyright 2012 Unknown <diogo@arch>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

"""
The `base` module includes the `Base` class, which is inherited by `Alignment`
and `AlignmentList` classes and provides several methods of general use.

It also defines the `CleanUp` decorator used by TriSeq and TriStats
to handle the generation of temporary data during their execution as well
as keyboard interruptions.
"""

try:
    from process.error_handling import InputError, EmptyAlignment
except ImportError:
    from trifusion.process.error_handling import InputError, EmptyAlignment

import os
import sys
import time
import shutil
import traceback
from collections import OrderedDict, Counter

dna_chars = ["a", "t", "g", "c"]

aminoacid_table = OrderedDict({"a": ["Alanine", "nonpolar", "neutral"],
                               "r": ["Arginine", "polar", "positive"],
                               "n": ["Asparagine", "polar", "neutral"],
                               "d": ["Aspartate", "polar", "negative"],
                               "c": ["Cysteine", "nonpolar", "neutral"],
                               "e": ["Glutamate", "polar", "negative"],
                               "q": ["Glutamine", "polar", "neutral"],
                               "g": ["Glycine", "nonpolar", "neutral"],
                               "h": ["Histidine", "polar", "neutral"],
                               "i": ["Isoleucine", "nonpolar", "neutral"],
                               "l": ["Leucine", "nonpolar", "neutral"],
                               "k": ["Lysine", "polar", "positive"],
                               "m": ["Methionine", "nonpolar", "neutral"],
                               "f": ["Phenylalanine", "nonpolar", "neutral"],
                               "p": ["Proline", "nonpolar", "neutral"],
                               "s": ["Serine", "polar", "neutral"],
                               "t": ["Threonine", "polar", "neutral"],
                               "w": ["Tryptophan", "nonpolar", "neutral"],
                               "y": ["Tyrosine", "polar", "neutral"],
                               "v": ["Valine", "nonpolar", "neutral"],
                               "u": ["Selenocysteine", "", ""],
                               "o": ["Pyrrolysine", "", ""],
                               "x": ["Missing", "", ""]})

iupac = {"ag": "r", "ct": "y", "cg": "s", "at": "w", "gt": "k", "ac": "m",
         "cgt": "b", "agt": "d", "act": "h", "acg": "v", "acgt": "n"}

iupac_rev = {"v": "acg", "r": "ag", "m": "ac", "s": "cg", "d": "agt",
             "b": "cgt", "n": "acgt", "h": "act", "y": "ct", "w": "at",
             "k": "gt"}

iupac_conv = {"v": "acg", "r": "ag", "m": "ac", "s": "cg", "d": "agt",
              "b": "cgt", "n": "acgt", "h": "act", "y": "ct", "w": "at",
              "k": "gt", "a": "aa", "t": "tt", "c": "cc", "g": "gg"}


class CleanUp(object):
    """Decorator class that handles temporary data for TriSeq and TriStats.

    This decorator class wraps the main execution functions of TriSeq and
    TriStats programs. The __init__ requires only the function reference,
    defines the name of the temporary directory and evaluates whether the
    provided `func` is from TriSeq or TriStats. The only requirement of
    `func` is that its first argument is the argparser namespace (that is,
    the arguments must be parsed in the corresponding program before
    calling the main function).
    
    Parameters
    ----------
    func : function
        Main function of TriSeq or TriStats
    
    Attributes
    ----------
    func : function
        Main function of TriSeq or TriStats
    temp_dir : str
        Path to temporary directory where the temporary data will be stored
    idx : int
        Integer identifier of the main function's program. 0 for TriSeq and
        2 for TriStats
        
    See Also
    --------
    print_col
    
    """

    def __init__(self, func):
        self.func = func
        # Set name of temporary directory
        self.temp_dir = ".trifusion-temp"
        # Evaluate whether func is from TriSeq or TriStats from its name
        self.idx = 0 if self.func.__name__ == "main_parser" else 2

    def __call__(self, *args):
        """Wraps the call of `func`.
        
        When the main `func` is called, this code is wrapped around its
        execution. This mainly ensures that the temporary data stored in
        `temp_dir` is correctly removed at the end of the execution, whether
        it terminals successfully or not. It also clocks the duration of the
        execution.
        
        Parameters
        ----------
        args : list
            Arbitrary list of positional arguments of `func`. The only
            requirement is that the first element is the argparse namespace
            object.

        """

        try:
            # Set starting time for clocking execution duration
            start_time = time.time()

            # Execute main function
            self.func(*args)

            # If program was not executed with 'quiet' flag, print final
            # execution message
            if not args[0].quiet:
                print_col("Program execution successfully completed in %s "
                          "seconds" %
                          (round(time.time() - start_time, 2)), GREEN,
                          self.idx)

        # Handle execution termination via Ctrl+C. Ensures that all temporary
        # data is remove.
        except KeyboardInterrupt:
            # Removing temporary directory, if any
            if os.path.exists(self.temp_dir):
                shutil.rmtree(self.temp_dir)
            print_col("Interrupting, by your command", RED, self.idx)

        # The broad exception handling is used to remove the temporary
        # directory under any circumstances
        except Exception:
            traceback.print_exc()
            # Removing temporary directory, if any
            if os.path.exists(self.temp_dir):
                shutil.rmtree(self.temp_dir)

            if not args[0].quiet:
                print_col("Program exited with errors!", RED, self.idx)

        # Removing temporary directory, if any
        if os.path.exists(self.temp_dir):
            shutil.rmtree(self.temp_dir)


def merger(ranges):
    """Generator that merges continuous ranges of tuples in a list.
    
    Parameters
    ----------
    ranges : list
        List of tuples, each with two integer elements defining a range,
        (0, 100) for instance.
    
    Example
    -------
    If the provide ranges are [(1, 234), (235, 456), (560, 607), (607,789)]
    this generator will yield the elements (1, 456) and (560, 789)
    """
    previous = 0
    last_start = 0
    for st, en in ranges:
        if not previous:
            last_start = st
            previous = en
        elif st - 1 == previous:
            previous = en
        else:
            yield last_start, previous
            previous = en
            last_start = st

    yield last_start, en


def has_colours(stream):
    if not hasattr(stream, "isatty"):
        return False
    if not stream.isatty():
        return False # auto color only on TTYs
    try:
        import curses
        curses.setupterm()
        return curses.tigetnum("colors") > 2
    except Exception as e:
        print(e.message)
        # guess false in case of error
        return False

# Support for terminal colors
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)
has_colours = has_colours(sys.stdout)


def print_col(text, color, i=0, quiet=False):
    """Custom print function for terminal updates of CLI programs.
    
    This print function homogenizes the progress logging of all CLI programs
    while providing some freedom on the formatting of the progress message,
    name the colors of the messages. A list of terminal colors is provided
    and a `has_colors` function checks whether they are supported on the
    terminal. The colors in use are green for normal logging, yellow for
    warnings and red for errors. The final formatting of the message is
    something like:
    
    [<Program Name>[-Error/Warning]] <message>
    
    Parameters
    ----------
    text : str
        The message that will appear in the terminal
    color : variable reference
        Reference to the terminal colors defined in process.base. Provided
        that they are imported in the module where they are being called, the
        options are: {GREEN, YELLOW, RED}
    i : int
        Integer that specified which CLI program is being logged. The options
        are:
        0: TriSeq
        1: OrthoMCL Pipeline
        2: TriStats
        3: TriOrtho
    quiet : bool
        Determines whether the message is logged. If True, no messages are
        printed to the terminal
    """
    if not quiet:
        p = ["TriSeq", "OrthoMCl Pipeline", "TriStats", "TriOrtho"]
        suf = {GREEN: "[%s] " % p[i], YELLOW: "[%s-Warning] " % p[i],
               RED: "[%s-Error] " % p[i]}
        if has_colours:
            seq = "\x1b[1;%dm" % (30 + color) + suf[color] + "\x1b[0m" + text
            print(seq)
        else:
            print(text)

    if color is RED:
        raise SystemExit(1)


class Base(object):

    def autofinder(self, reference_file):
        """Autodetects format, missing data symbol and sequence type.
        
        Attempts to find the file format, missing data symbol and sequence
        type from a reference file. Due to performance reasons, this function
        will only read the `referenced_file` until it finds the first sequence.
        Then, it evaluates the file format and sequence type. It also
        attempts to get the missing data symbol, but the first sequence may
        have no missing data. In that case, the missing data symbol will remain
        undefined and will be re-evaluated during alignment parsing.
        
        Parameters
        ----------
        reference_file : str
            Path to sequence file
        
        Returns
        -------
        fmt : str
            File format of `reference_file`
        code : tuple
            The sequence type as string in first element and missing data
            symbol as string in the second element
        
        See Also
        --------
        guess_code
        
        """

        file_handle = open(reference_file, "r")

        # Set to True when the format has been detected
        format_found = False

        # If input file is not a simple text file, which means it's invalid,
        # handle this exception
        try:
            header = file_handle.readline()
        except UnicodeDecodeError:
            return InputError("Invalid input file.")

        try:
            # Skips first empty lines, if any
            while header.startswith("\n"):
                header = next(file_handle)

            # Recognition of NEXUS files is based on the existence of the
            # string "#NEXUS" in the first non-empty line
            if header.upper().strip().startswith("#NEXUS"):
                fmt = "nexus"
                format_found = True
                while True:
                        line = next(file_handle)
                        if line.strip().lower() == "matrix":
                            next_line = next(file_handle)
                            while next_line.startswith("\n"):
                                next_line = next(file_handle)
                            sequence = "".join(next_line.split()[1:]).strip()
                            break

            # Recognition of Stockholm files is based on the existence of the
            # string "# stockholm" in the first non-empty line
            # (case insensitive)
            elif header.upper().strip().startswith("# STOCKHOLM") or \
                    header.upper().strip().startswith("#STOCKHOLM"):
                fmt = "stockholm"
                format_found = True
                while True:
                    line = file_handle.readline()
                    if not line.startswith("#") and line.strip() != "":
                        sequence = line.split()[1]
                        break

            # Recognition of FASTA or .loci files is based on the existence
            # of a ">" character as the first character of a non-empty line
            elif header.strip().startswith(">"):
                next_line = next(file_handle)
                if next_line.strip().startswith(">"):
                    fmt = "loci"
                    format_found = True
                    sequence = header.split()[-1].strip()
                else:
                    fmt = "fasta"
                    format_found = True
                    sequence = next_line.strip()
                    for line in file_handle:
                        if line.strip() != "" and line.strip()[0] != ">":
                            sequence += line.strip()
                        elif line.strip() != "" and line.strip()[0] == ">":
                            break

            # Recognition of Phylip files is based on the existence of two
            # integers separated by whitespace on the first non-empy line
            elif len(header.strip().split()) == 2 and header.strip().split()[0]\
                    .isdigit() and header.strip().split()[1].isdigit():

                fmt = "phylip"
                format_found = True
                sequence = "".join(file_handle.readline().split()[1:]).strip()

            # Recognition of ipyrad loci file, which does not start with a ">"
            # character
            if not format_found:
                while not header.startswith("//"):
                    try:
                        header = next(file_handle)
                    except StopIteration:
                        break
                    if header.startswith("//"):
                        if header.count("|") == 2:
                            fmt = "loci"
                            format_found = True
                            sequence = next(file_handle).split()[1]

            # Check if there is any sequence. If not, the alignment file has no
            # sequence
            if not format_found:
                return InputError("Unknown input file format.")
            if sequence.replace("-", "") == "":
                return EmptyAlignment("Alignment is empty")

            # Guessing the genetic code
            code = self.guess_code(sequence)

        except StopIteration:
            return EmptyAlignment("Alignment is empty")

        return fmt, code

    @staticmethod
    def get_loci_taxa(loci_file):
        """
        Get the list of taxa from a .loci file.
        
        This is required prior to parsing the alignment in order to
        correctly add missing data when certain taxa are not present in
        a locus
        
        Parameters
        ----------
        loci_file : str
            Path to the .loci file.
        
        Returns
        -------
        taxa_list : list
            List of the taxon names as strings.
        
        
        """

        file_handle = open(loci_file)
        taxa_list = []

        for line in file_handle:
            # Skip empty lines and lines starting with //
            # The .lstrip(">") supports the new format of pyRAD's .loci
            # where taxon names start with a ">" character
            if not line.strip().startswith("//") and line.strip() != "":
                taxon = line.strip().split()[0].lstrip(">")
                if taxon not in taxa_list:
                    taxa_list.append(taxon)

        return taxa_list

    @staticmethod
    def guess_code(sequence):
        """
        Guess the sequence type, i.e. protein.
        
        Guesses the code of the provided `sequence`, that is, if it is a
        DNA or Protein sequence based on the first sequence of
        the reference file. The second item of the returned `code` variable
        is a placeholder for the missing data symbol (can be either "?", "n"
        or "x"). This symbol will be evaluated during the Alignment parsing.

        Parameters
        ----------
        sequence : string
            Sequence string that will be used to guess the type
        
        Returns
        -------
        code : list
            Provides the (<sequence type>, <missing symbol character>).
            sequence type can be either "DNA" or "Protein". The
            missing symbol character may be "?", "n", "x" or None (See Notes).
        
        See also
        --------
        autofinder
        process.sequence.Alignment.read_alignment
        
        Notes
        -----
        The missing symbol character that is provided in the returned `code`
        tuple may be None if the missing character cannot be determined
        from the first reference sequence that is used to guess the
        sequence type. If none of the potential missing data symbol is
        present in this reference sequence, it will remain undetermined and
        will be evaluated during the alignment parsing.
        
        """

        # Removes gaps from the sequence so that the frequencies are not biased
        sequence = sequence.upper().replace("-", "")

        # Try to evaluate whether missing data is in ? or N/X format. Since
        # the '?' character should only occur in sequences as missing data
        # (unlike the N character, which can be an amino acid residue),
        # we assume that if it exists in the sequence, that's the missing
        # data character
        if sequence.count("?"):
            missing = "?"
        else:
            missing = None

        dna_count = sequence.count("A") + sequence.count("T") + \
            sequence.count("G") + sequence.count("C") + \
            sequence.count("N")

        dna_proportion = float(dna_count) / float(len(sequence))

        # The 0.9 cut-off has been effective so far
        if dna_proportion > 0.9:
            if sequence.count("N"):
                missing = "n"
            code = ["DNA", missing]
        else:
            if sequence.count("X"):
                missing = "x"
            code = ["Protein", missing]
        return code

    @staticmethod
    def rm_illegal(taxon_string):
        """
        Removes illegal characters from taxon name.
        
        The illegal characters are defined in the `illegal_chars`
        variable
        
        Parameters
        ----------
        taxon_string : str
            Taxon name
        
        Returns
        -------
        clean_name : str
            Taxon name without illegal characters
        
        """

        # Additional illegal characters are added here
        illegal_chars = [":", ",", ")", "(", ";", "[", "]", """, """]

        clean_name = "".join([char for char in taxon_string if char not in
                              illegal_chars])

        return clean_name

    @staticmethod
    def duplicate_taxa(taxa_list):
        """
        Identified duplicate items in a list.
        
        Used to detect, for instance, duplicated taxa in the Alignment object
        
        Parameters
        ----------
        taxa_list : list
            List with taxon names
        
        Returns
        -------
        duplicated_taxa : list
            List with duplicated taxa from `taxa_list`
        """

        duplicated_taxa = [x for x, y in Counter(taxa_list).items()
                           if y > 1]

        return duplicated_taxa

    @staticmethod
    def read_basic_csv(file_handle):
        """
        Reads a basic CSV into a list.
        
        Parses a simples CSV file with only one column and one
        or more lines while stripping whitespace.
        
        Parameters
        ----------
        file_handle : file object
            File object of the CSV file
        
        Returns
        -------
        result : list
            Contents of the CSV file with each line in a different entry
        
        """

        result = []

        for line in file_handle:

            result.append(line.strip())

        return result


__author__ = "Diogo N. Silva"



#!/usr/bin/python2
# -*- coding: utf-8 -*-

import sys

reload(sys)
sys.setdefaultencoding("utf8")

def gui_exec():

    try:
        from trifusion.app import main
    except ImportError:
        from app import main

    main()


if __name__ == "__main__":

    import multiprocessing

    multiprocessing.freeze_support()

    gui_exec()



import os
from os.path import join

from memory_profiler import memory_usage
import traceback

from process.sequence import AlignmentList
import process.data as data

data_dir = "/home/diogo/Diogo/Science/Scripts/packages/TriFusion_benchmarks/data/"
db_path = "/home/diogo/Diogo/TriFusion_testdata/benchmark.db"
bench_dir = "/home/diogo/Diogo/TriFusion_testdata/benchmarks_out"

res = []

# Uncomment active data set
# ds_path = join(data_dir, "614_protein")
# ds_path = join(data_dir, "3093_protein")
# ds_path = join(data_dir, "3093_141_protein")
# ds_path = join(data_dir, "3093_376_protein")
# ds_path = join(data_dir, "7378_dna")
# ds_path = join(data_dir, "1_40_large")
# ds_path = join(data_dir, "1_40_large_interleave")
# ds_path = join(data_dir, "1_40_large_interleave_nex")
# ds_path = join(data_dir, "1_376_large")
# ds_path = join(data_dir, "2_big_data")
ds_path = join(data_dir, "52285_loci")

#mfilter = ["Postia_placenta"]
#mfilter = ["535_12"]
mfilter = ["Ang_30"]

ds = [join(ds_path, x) for x in os.listdir(ds_path)]

# rev_ds = join(data_dir, "614_protein_rev")
# rev_ds = join(data_dir, "3093_protein_rev")
# rev_ds = join(data_dir, "3093_141_protein_rev")
# rev_ds = join(data_dir, "3093_376_protein_rev")
# rev_ds = join(data_dir, "7378_dna_rev")
# rev_ds = join(data_dir, "2_big_data_rev")
# rev_ds = join(data_dir, "52285_loci_rev")

# ds = [join(rev_ds, x) for x in os.listdir(rev_ds) if x.endswith(".phy")]
# part_file = [join(rev_ds, x) for x in os.listdir(rev_ds) if not x.endswith(".phy")][0]

# @profile
def read_ds(ds):
    print("Profiling 'read_ds' with a data set of {} files".format(len(ds)))
    x = AlignmentList(ds, sql_db=db_path)

# @profile
def convert_ds(aln_obj):

    aln_obj.write_to_file(["fasta"], output_dir=bench_dir)

# @profile
def convert_ds_int(aln_obj):
    print("Profiling 'convert_ds_int' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.write_to_file(["nexus"], output_dir=bench_dir, interleave=True)

# @profile
def concatenate_ds(aln_obj):
    print("Profiling 'concatenate_ds' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.concatenate()

def collapse_ds(aln_obj):
    print("Profiling 'collapse_ds' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.collapse(use_main_table=True, dest=bench_dir)

def consensus_ds(aln_obj):
    print("Profiling 'consensus_ds' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.consensus("Soft mask", single_file=True, use_main_table=True)

def taxa_contain_filter(aln_obj):
    print("Profiling 'taxa_contain_filter' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.filter_by_taxa(mfilter, "contain")

def min_taxa_filter(aln_obj):
    print("Profiling 'min_taxa_filter' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.filter_min_taxa(80)

def missing_data_filter(aln_obj):
    print("Profiling 'missing_data_filter' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.filter_missing_data(50, 75)

def codon_filter(aln_obj):
    print("Profiling 'codon_filter' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.filter_codon_positions([True, True, False])

def variable_filter(aln_obj):
    print("Profiling 'variable_filter' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.filter_segregating_sites(0, 200)

def informative_filter(aln_obj):
    print("Profiling 'informative_filter' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.filter_informative_sites(0, 200)

def code_gaps(aln_obj):
    print("Profiling 'code_gaps' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    aln_obj.code_gaps()

def reverse_concat(aln_obj, part_file):
    print("Profiling 'reverse_concat' with a data set of {} files and "
          "{} taxa".format(
        len(aln_obj.alignments), len(aln_obj.taxa_names)))
    partition = data.Partitions()
    partition.read_from_file(part_file)
    saln = aln_obj.alignments.values()[0]
    saln.set_partitions(partition)
    aln_obj.set_partition_from_alignment(saln)
    aln_obj.reverse_concatenate()

    # Post reverse ops
    print("Writing files")
    aln_obj.write_to_file(["nexus"], output_dir=bench_dir)

# Always start by reseting database
if os.path.exists(db_path):
    os.remove(db_path)

aln_obj = AlignmentList(ds, sql_db=db_path)

try:
    control_mem = memory_usage()
    # print("Control memory usage at: %s" % round(max(control_mem), 1))
    # mem_usage = memory_usage((read_ds, (ds, )))
    # mem_usage = memory_usage((convert_ds, (aln_obj,)))
    # mem_usage = memory_usage((convert_ds_int, (aln_obj, )))
    # mem_usage = memory_usage((concatenate_ds, (aln_obj,)))
    # mem_usage = memory_usage((collapse_ds, (aln_obj, )))
    # mem_usage = memory_usage((consensus_ds, (aln_obj, )))
    #mem_usage = memory_usage((taxa_contain_filter, (aln_obj, )))
    # mem_usage = memory_usage((min_taxa_filter, (aln_obj, )))
    # mem_usage = memory_usage((missing_data_filter, (aln_obj, )))
    # mem_usage = memory_usage((codon_filter, (aln_obj, )))
    # mem_usage = memory_usage((variable_filter, (aln_obj, )))
    # mem_usage = memory_usage((informative_filter, (aln_obj, )))
    # mem_usage = memory_usage((reverse_concat, (aln_obj, part_file, )))
    mem_usage = memory_usage((code_gaps, (aln_obj, )))
    print("Maximum memory usage: %s" % round(max(mem_usage) - max(control_mem), 1))
except Exception as e:
    print(e)
    traceback.print_exc()

# os.remove(db_path)